// TODO
/* --------------------------------------------------
 * RASP24
 * デバイスバグ
 *
 * WsSetDaGain
 * WsSetDaLoopBackChannel
 * の戻り値が常に1になっている。
 * これらの関数を使用した場合の動作
 * o DAゲイン設定
 * o ループバックモード設定
 * の動作は未確認。
 *
 * ファーム、FPGAを最新のものに更新後に要確認
 * -------------------------------------------------- */
// TODO

// FIXME
/* --------------------------------------------------
 * 以下はバグでうまく動かないので要修正
 * ゲイン設定 => 修正済み
 * ループバック設定
 *
 * 2012/02/09:
 *   rasp で同期モードで使ったとき、ADとDAのゲインが
 *   どちらも設定できないバグを修正
 * -------------------------------------------------- */
// FIXME

#include "rasp.h"
#include <limits.h>

#define PERIODSIZE 1023
#define CHANNELS_PER_BOARD 9
#define MAX_BOARDS 16
#define MAX_DA_CHANNELS 2
#define MAX_AD_CHANNELS(nboards) ((nboards)*9)
#define MIN_RASP_BUFFER_SIZE (1<<15)
#define MAX_RASP_BUFFER_SIZE (1<<20)
#define DA_UNKNOWN_MODE INT_MAX
#define DATA_FORMAT_UNKNOWN INT_MAX
#define NOT_USE_VAL INT_MAX

/* --------------------------------------------------
 * RASP24
 * ゲイン設定値
 *
 * デバイスには
 * DA: [-63, 0]
 * AD: [0, 4]
 * の範囲で設定が可能
 * -------------------------------------------------- */
#define DA_GAIN_MIN (-63)
#define DA_GAIN_MAX    0
#define AD_GAIN_MIN    0
#define AD_GAIN_MAX    4

typedef int rasp_pcm_format_t;

typedef struct {
	WS_HANDLE_t * device;
	wios_buffer_t buffer;
	char * ip_addr;
} rasp_priv_t;

/* --------------------------------------------------
 * RASP24
 * 対応WAVフォーマット	RASP設定値						 	プログラム上で扱うデータ
 *
 * Signed 16bit PCM			DATA_FORMAT_16BIT_PCM		int16_t
 *   (RASPでは下位8bitに0を付加した24bitデータとして扱われる)
 *
 * Signed 24bit PCM			DATA_FORMAT_32BIT_PCM		int24_t (typedef int32_t int24_t)
 *   (RASPでは下位24bitのみが使用され、24bitデータとして扱われる)
 * -------------------------------------------------- */
static const rasp_pcm_format_t rasp_format[] = {
	DATA_FORMAT_UNKNOWN,   /* パディング     */
	DATA_FORMAT_16BIT_PCM, /* 16bit/item WAV */
	DATA_FORMAT_32BIT_PCM, /* 24bit/item WAV */
	DATA_FORMAT_UNKNOWN
};

static int supported_rate[] = {
	11025,
	16000,
	22050,
	32000,
	44100,
	48000,
};

// FIXME need to fix BUG!
/* チャンネル数(添字)をRASPマクロに */
static int channel_mode[] = {
	NOT_USE_VAL,    /* 0: パディング */
	DA_SINGLE_MODE, /* 1: 1ch */
	DA_STEREO_MODE, /* 2: 2ch or ループバック */
	DA_UNKNOWN_MODE /* 3: 未対応 */
};

static int channel_flag[] = {
	AD_CH0_CAPTURE ,
	AD_CH1_CAPTURE ,
	AD_CH2_CAPTURE ,
	AD_CH3_CAPTURE ,
	AD_CH4_CAPTURE ,
	AD_CH5_CAPTURE ,
	AD_CH6_CAPTURE ,
	AD_CH7_CAPTURE ,
	AD_CH8_CAPTURE ,
	AD_CH9_CAPTURE ,
	AD_CH10_CAPTURE,
	AD_CH11_CAPTURE,
	AD_CH12_CAPTURE,
	AD_CH13_CAPTURE,
	AD_CH14_CAPTURE,
	AD_CH15_CAPTURE,
	-1
};



/* STATIC FUNCTION PROTOTYPE */
/* -------------------------------------------------- */

static int check_sampling_rate(int rate);
static int check_ad_gain(int gain);
static int check_da_gain(int gain);

static int priv_init(wios_file_t * ft);
static int device_write(wios_file_t * ft, wios_buffer_t * buf, int len);
static int device_read(wios_file_t * ft, wios_buffer_t * buf, int len);
static char * rasp_buffer_malloc(wios_file_t * ft, wios_buffer_t * buf, int first_nframes);

static int get_numlist(int * gain, char * s, int channels);
static int rasp_set_da_gain(wios_file_t * ft);
static int rasp_set_ad_gain(wios_file_t * ft);
static int rasp_set_da_gain_all(wios_file_t * ft);
static int rasp_set_ad_gain_all(wios_file_t * ft);
static int rasp_set_loopback_ch(wios_file_t * ft);



/* HANDLER FUNCTION */
/* -------------------------------------------------- */

/* --------------------------------------------------
 * RASP24
 * RASPのCH数モードについてのデバイス設定値
 *
 * DAのCH数モード			PCMフォーマット		備考
 * DA_SINGLE_MODE			1ch								
 * DA_STEREO_MODE			2ch								ループバックの際はこのモード
 * DA_MONORAL_MODE		(ボード枚数)ch		ボード毎に異なるデータを出力する	
 * -------------------------------------------------- */

/* --------------------------------------------------
 * RASP24
 * 同期時のデバイス設定について
 * 
 * 再録同期時はデバイスにはADとDAについて
 * チャンネル数、エンコーディング
 * を異なる値に設定することができる。サンプリングレートについては
 * 異なる値に設定した場合にデバイスでうまく再録が行われない。
 * -------------------------------------------------- */
static int device_init(wios_file_t * ft)
{
	rasp_priv_t    * priv = ft->priv;
	wios_fmtinfo_t * fmt = &ft->fmtinfo;
	int header_flag = 0; /* RASPヘッダをなしに設定する */
	int status;

	/* デバイスハンドル作成 */
	if (ft->io_type == wios_da) {
		if ((priv->device = WsDaHandleOpen(priv->ip_addr)) == NULL)
			WIOS_ERROR("cannot allocate hardware parameter strucutre", NULL);
	} else {
		if ((priv->device = WsHandleOpen(priv->ip_addr)) == NULL)
			WIOS_ERROR("cannot allocate hardware parameter strucutre", NULL);
	}

	/* デバイスハンドルを設定 */
	ft->dp = priv->device;

	/* サンプリングレートを設定 */
	if (check_sampling_rate(ft->fmtinfo.rate) == WIOS_EOF)
		WIOS_ERROR("cannot set sampling rate `%d'", ft->fmtinfo.rate);
	WsSetSamplingFreq(priv->device, fmt->rate);

	/* 量子化ビット数を設定 */
	if (rasp_format[fmt->format] == DATA_FORMAT_UNKNOWN)
		WIOS_ERROR("cannot set format `%s'", wios_format_to_string(fmt->format));
	if (ft->io_type == wios_da) {
		if (WsSetDaDataFormat(priv->device, rasp_format[fmt->format]) < 0)
			WIOS_ERROR("cannot set format `%s'", wios_format_to_string(fmt->format));
	} else {
		if (WsSetAdDataFormat(priv->device, rasp_format[fmt->format]) < 0)
			WIOS_ERROR("cannot set format `%s'", wios_format_to_string(fmt->format));
	}

	// TODO デバイス更新後要確認
	/* ループバックモードの設定 */
	if (ft->io_type == wios_da) {
		if (rasp_da_lb_ch_str && rasp_ad_lb_ch_str)
			lb_flag = WIOS_TRUE;

		if (lb_flag) {
			/* ループバックを設定する */
			/* デバイスバグの為未確認 */
			if (rasp_set_loopback_ch(ft) == WIOS_EOF)
				goto error;

			/* ループバックの時はチャンネルををステレオモードに設定 */
			if ((status = WsSetDaChannelMode(priv->device, DA_STEREO_MODE)) != 0)
				WIOS_ERROR("cannot set DA stereo mode (status:0x%08x))", status);
		}
	}
	// TODO
 
	/* チャンネル数を設定 */
	if (ft->io_type == wios_da) {
		int need_channel_num;
		if (!lb_flag) {
			if (!da_monoral_flag && fmt->channels > MAX_DA_CHANNELS)
				WIOS_ERROR("cannot set channel count `%d'", fmt->channels);
			if (da_monoral_flag) {
				if ((status = WsSetDaChannelMode(priv->device, DA_MONORAL_MODE)) != 0)
					WIOS_ERROR("cannot set DA monoral mode (status:0x%08x)", status);
			} else {
				// FIXME need to fix BUG!
				if (WsSetDaChannelMode(priv->device, channel_mode[fmt->channels]) != 0)
					WIOS_ERROR("cannot set channel count `%d'", fmt->channels);
			}
		}

		/* 再生データのチャンネル数が正しいかをチェック */
		if ((status = WsGetDaChannelNum(priv->device, &need_channel_num)) != 0)
			WIOS_ERROR("cannot get channel count (status:0x%08x)", status);
		if (need_channel_num != fmt->channels)
			WIOS_ERROR("playback cannel count must be %d", need_channel_num);

	} else {
		int i, j, board_num, need_board_num;
		/* ボードの枚数が足りているかチェック */
		if (WsGetBoardNum(priv->device, &board_num) < 0)
			WIOS_ERROR("cannot get the number of boards", NULL);
		if (fmt->channels > MAX_AD_CHANNELS(board_num))
			WIOS_ERROR("cannot capture `%d' channels (only %d boards)", fmt->channels, board_num);
		need_board_num = fmt->channels/CHANNELS_PER_BOARD + (fmt->channels%CHANNELS_PER_BOARD?1:0);

		/* チャンネルを設定 */
		for (i = 0; i < need_board_num; i++) {
			int channels = 0;
			for (j = 0; j < CHANNELS_PER_BOARD && j+i*CHANNELS_PER_BOARD < fmt->channels; j++)
				channels |= channel_flag[j];
			WsSetCaptureChannels(priv->device, i, channels);
		}
		for (i = need_board_num; i < board_num; i++) {
			int channels = 0;
			WsSetCaptureChannels(priv->device, i, channels);
		}
	}

	/* DAとADのゲインを設定 */
	if (ft->io_type == wios_da)
	{
		if (wios_mode == wios_play || wios_mode == wios_sync)
		{
			if (rasp_da_gain_str && rasp_set_da_gain(ft) == WIOS_EOF)
				goto error;
			if (rasp_set_da_gain_all(ft) == WIOS_EOF)
			goto error;
		}
	}
	if (ft->io_type == wios_ad)
	{
		if (wios_mode == wios_rec || wios_mode == wios_sync)
		{
			if (rasp_ad_gain_str && rasp_set_ad_gain(ft) == WIOS_EOF)
				goto error;
			if (rasp_set_ad_gain_all(ft) == WIOS_EOF)
				goto error;
		}
	}

	/* A/Dデータにヘッダ情報を付加しないように設定 */
	if (ft->io_type == wios_ad) {
		if (WsSetHeaderFlag(priv->device, header_flag) < 0)
			WIOS_ERROR("cannot set header flag `%d'", header_flag);
	}

	return WIOS_SUCCESS;

error:
	return WIOS_EOF;
}

static wios_file_t * open_read(
	const char             * path,
	const char             * dev,
	const wios_execinfo_t  * info)
{
	return open_read_(path, dev, info, priv_init);
}

static wios_file_t * open_write(
		const char             * path,
		const char             * dev,
		const wios_execinfo_t  * info,
		const wios_fmtinfo_t   * fmt)
{
	/* RASPの場合は特別にperiodsizeを設定 */
	wios_execinfo_t * rasp_info = (wios_execinfo_t *)info;
	rasp_info-> periodsize = PERIODSIZE;
	return open_write_(path, dev, info, fmt, priv_init);
}

/* --------------------------------------------------
 * RASP24
 * 再生準備
 *
 * RASPは最大1MB分のデータをバッファ可能。
 * リファレンスにはD/A開始前に32kB以上のデータがないとD/Aは開始しないとあるが、
 * 自分の環境では64kB程度のデータを書きこまないとD/Aが開始されなかったため、
 * 64kB以上書き込んでから再生するようにしてある。
 * -------------------------------------------------- */
static void * playback(void * arg)
{
	/* 再生準備が終わるまで呼び出し元をロック */
	pthread_mutex_lock(&ready_mutex);
	wios_file_t    * ft = (wios_file_t *)arg;
	rasp_priv_t    * priv = ft->priv;
	wios_buffer_t  * buf = &priv->buffer;
	wios_signal_t  * sig = &ft->signal;

	int frames;
	int done = 0;
	int len = ft->signal.execinfo.periodsize;
	int req_nframes;

	int first_nframes;
	int status, i;
	int byte_per_periodsize = ft->signal.byte_per_item * ft->fmtinfo.channels;

	if (ft->signal.execinfo.sec == WIOS_ALL_SEC)
		req_nframes = ft->signal.all_frames;
	else
		req_nframes = ft->fmtinfo.rate * ft->signal.execinfo.sec;

	for (first_nframes = 0; first_nframes*byte_per_periodsize < MIN_RASP_BUFFER_SIZE*2; first_nframes++)
		if (first_nframes >= ft->signal.all_frames) break;
	
	if (first_nframes*byte_per_periodsize < MIN_RASP_BUFFER_SIZE*2)
		WIOS_ERROR("duration of DA file is too short", NULL);

	if (rasp_buffer_malloc(ft, buf, first_nframes) == NULL)
		goto error;

	/* first_nframesフレーム分のデータを先に読み込んでおく */
	frames = lsf_file_read(ft, buf->fl, first_nframes);
	/* first_nframesフレーム分のデータを先に書き込んでおく */
	device_write(ft, buf, frames);
	done += frames;
	/* 次のデータを読み込んでおく */
	len = min(len, ft->signal.all_frames-done);
	frames = lsf_file_read(ft, buf->fl, len);

	/* 再録同期時のDA準備 */
	if (wios_mode == wios_sync && !use_2dev_option_flag) {
		if ((status = WsDaReady(ft->dp)) != 0)
			WIOS_ERROR("cannot ready playback (status:0x%08x)", status);
	}

	/* 再生準備完了をメインスレッドに知らせる */
	pthread_cond_signal(&ready_cond);
	pthread_mutex_unlock(&ready_mutex);

	if (wios_mode == wios_sync) {
		/* メインスレッドで同期を開始させるまで待機 */
		pthread_mutex_lock(&sync_mutex);
		while (!sync_start_flag)
			pthread_cond_wait(&sync_cond, &sync_mutex);
		pthread_mutex_unlock(&sync_mutex);
	}

	/* 以下同期スタート */
	if (print_mesg_flag)
		fprintf(stdout, "Start playback (%s)\n", ft->file_name);
  if (print_time_flag)
    fprintf(stdout, "Playback_start_Time:%lf\n", wios_get_time());

	if (wios_mode == wios_play || use_2dev_option_flag) {
		if ((status = WsDaStart(ft->dp)) != 0)
		WIOS_ERROR("cannot start playback (status:0x%08x)", status);
	}
	
	/* 再生ループ */
	while (1) {
		/* SIGINTを受け取っていたらブレーク */
		if (loop_break_flag) break;

		device_write(ft, buf, frames);

		done += frames;

		if (done >= req_nframes) break;

		len = min(len, req_nframes-done);

		frames = lsf_file_read(ft, buf->fl, len);

		/* ファイル読み取りエラー */
		if (frames <= 0) break;
	}

	/* 再生終了を待つ */
	WsDaStopWait(ft->dp);
	WsDaStop(ft->dp);

	/* 再生終了を録音スレッドに知らせる */
	playback_end_flag = WIOS_TRUE;

	if (print_mesg_flag)
		fprintf(stdout, "End playback (%s)\n", ft->file_name);

	wios_buffer_free(buf);

	return NULL;

error:
	if (pthread_mutex_destroy(&ready_mutex) == EBUSY)
		pthread_mutex_unlock(&ready_mutex);

	wios_buffer_free(buf);

	return NULL;
}

/* --------------------------------------------------
 * RASP24
 * キャプチャーに関しての注意
 *  
 * periodsizeはAPIに1023[frame]で固定されている。
 * APIはこの値を仮定して実装されているため
 * periodsizeは1023[frame]に固定してある。
 * -------------------------------------------------- */
static void * capture(void * arg)
{
	/* 録音準備が終わるまで呼び出し元をロック */
	pthread_mutex_lock(&ready_mutex);
	wios_file_t * ft = (wios_file_t *)arg;
	rasp_priv_t * priv = ft->priv;
	wios_buffer_t * buf = &priv->buffer;
	wios_signal_t * sig = &ft->signal;

	int frames;
	int done = 0;
	int len = PERIODSIZE;
	int req_nframes;

	int written = 0;

	int status;

	if (wios_mode == wios_sync && !use_2dev_option_flag) {
		if (ft->signal.execinfo.sec == WIOS_ALL_SEC)
			req_nframes = ftda->signal.all_frames;
		else
			req_nframes = ft->fmtinfo.rate * ft->signal.execinfo.sec;
	} else {
		req_nframes = ft->fmtinfo.rate * ft->signal.execinfo.sec;
	}

	if (rasp_buffer_malloc(ft, buf, ft->signal.buf_size) == NULL)
		goto error;

	/* 録音準備完了をメインスレッドに知らせる */
	pthread_cond_signal(&ready_cond);
	pthread_mutex_unlock(&ready_mutex);

	if (wios_mode == wios_sync) {
		/* メインスレッドで同期を開始させるまで待機 */
		pthread_mutex_lock(&sync_mutex);
		while (!sync_start_flag)
			pthread_cond_wait(&sync_cond, &sync_mutex);
		pthread_mutex_unlock(&sync_mutex);
	}

	/* 以下同期スタート */
	if (print_mesg_flag)
		fprintf(stdout, "Start capture (%s)\n", ft->file_name);
  if (print_time_flag)
    fprintf(stdout, "Capture_start_Time:%lf\n", wios_get_time());
	
	if (wios_mode == wios_rec || use_2dev_option_flag) {
		if ((status = WsADStart(ft->dp)) != 0)
			WIOS_ERROR("cannot start playback (status:0x%08x)", status);
	} else {
		if ((status = WsAdDaStart(ft->dp)) != 0)
			WIOS_ERROR("cannot start playback and capure (status:0x%08x)", status);
	}

	/* 録音ループ */
	while (1) {
		/* SIGINTを受け取っていたらブレーク */
		if (loop_break_flag) break;

		frames = device_read(ft, buf, len);

		if (wios_mode != wios_rec || ft->signal.execinfo.sec != WIOS_ALL_SEC) {
			/* 余分なフレームを書きこまないようにする */
			if (frames > req_nframes-done) {
				if (ft->fmtinfo.format == WIOS_PCM_FORMAT_S24)
					ft->success_byte_cnt -= (frames-(req_nframes-done))*3*ft->fmtinfo.channels;
				else 
					ft->success_byte_cnt -= (frames-(req_nframes-done))*ft->signal.byte_per_item*ft->fmtinfo.channels;
				frames = req_nframes-done;
			}
		}

		lsf_file_write(ft, buf->fl, frames);

		done += frames;

		/* 録音モードで秒数が指定されなかった場合はブレークしない */
		if (wios_mode != wios_rec || ft->signal.execinfo.sec != WIOS_ALL_SEC)
			if (done >= req_nframes) break;
	}

	WsADStop(ft->dp);

	if (print_mesg_flag)
		fprintf(stdout, "End capture (%s)\n", ft->file_name);

	wios_buffer_free(buf);

	return NULL;

error:
	if (pthread_mutex_destroy(&ready_mutex) == EBUSY)
		pthread_mutex_unlock(&ready_mutex);

	wios_buffer_free(buf);

	return NULL;
}

static void finish(wios_file_t * ft)
{
	rasp_priv_t * priv = ft->priv;
	if (ft) {
		if (ft->fp) {
			sf_close(ft->fp);
			ft->fp = NULL;
		}
		if (ft->dp) {
			WsHandleClose(ft->dp);
			ft->dp = NULL;
		}
		if (priv) {
			free(priv);
			priv = NULL;
		}
		if (ft) {
			free(ft);
			ft = NULL;
		}
	}
}



/* STATIC FUNCTION */
/* -------------------------------------------------- */

static int check_sampling_rate(int rate)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(supported_rate); i++)
		if (supported_rate[i] == rate) return i;

	return WIOS_EOF;
}

static int check_ad_gain(int gain)
{
	if (gain >= AD_GAIN_MIN && gain <= AD_GAIN_MAX)
		return WIOS_SUCCESS;
	
	return WIOS_EOF;
}

static int check_da_gain(int gain)
{
	if (gain >= DA_GAIN_MIN && gain <= DA_GAIN_MAX)
		return WIOS_SUCCESS;
	
	return WIOS_EOF;
}

static int priv_init(wios_file_t * ft)
{
	rasp_priv_t * priv;

	if ((ft->priv = malloc(sizeof(rasp_priv_t))) == NULL)
		goto error;
	priv = ft->priv;

	/* RASP仕様時はデバイス名をIPアドレスをとする */
	priv->ip_addr = (char *)ft->dev_name;

	priv->buffer.fl = NULL;
	priv->buffer.dv = NULL;

	return WIOS_SUCCESS;

error:
	return WIOS_EOF;
}

static int device_write(wios_file_t * ft, wios_buffer_t * buf, int nframes)
{
	int i, err;
	int n = ft->signal.byte_per_item*ft->fmtinfo.channels;

	switch (ft->fmtinfo.format) {
		case WIOS_PCM_FORMAT_S16: {
			int16_t * buf1 = (int16_t *)buf->dv;
			for (i = 0; i < nframes*ft->fmtinfo.channels; i++) {
				buf1[i] = WIOS_CONVERT_32_TO_16BIT(buf->fl[i]);
				DA_CLIP_WARN(WIOS_16BIT_CLIP_COUNT(buf1[i], ft->clips));
				WIOS_RENEW_MIN_MAX_FRAME(buf1[i], ft->min_item_val, ft->max_item_val);
			}
			break;
		}
		case WIOS_PCM_FORMAT_S24: {
			int24_t * buf1 = (int24_t *)buf->dv;
			for (i = 0; i < nframes*ft->fmtinfo.channels; i++) {
				buf1[i] = WIOS_CONVERT_32_TO_24BIT(buf->fl[i]);
				DA_CLIP_WARN(WIOS_24BIT_CLIP_COUNT(buf1[i], ft->clips));
				WIOS_RENEW_MIN_MAX_FRAME(buf1[i], ft->min_item_val, ft->max_item_val);
			}
			break;
		}
		default: {
			break;
		}
	}

	/* nframes*n[byte]のデータをデバイスに書き込む */
	err = WsDaDataWrite(ft->dp, buf->dv, nframes*n);

	if (ft->fmtinfo.format == WIOS_PCM_FORMAT_S24) {
		if (err == nframes*n) {
			ft->success_byte_cnt += err > 0 ? (err*3)/4 : 0;
		} else {
			ft->success_byte_cnt += err > 0 ? (err*3)/4 : 0;
			ft->missing_byte_cnt += (nframes*n-err) > 0 ? (nframes*n-err*3)/4 : 0;
		}
	} else {
		if (err == nframes*n) {
			ft->success_byte_cnt += err > 0 ? err : 0;
		} else {
			ft->success_byte_cnt += err > 0 ? err : 0;
			ft->missing_byte_cnt += (nframes*n-err) > 0 ? (nframes*n-err) : 0;
		}
	}

	return err;
}

static int device_read(wios_file_t * ft, wios_buffer_t * buf, int nframes)
{
	int i, err;
	int n = ft->signal.byte_per_item*ft->fmtinfo.channels;

	/* nframes*n[byte]のデータをデバイスから読み込む */
	err = WsFrameRead(ft->dp, buf->dv, nframes*n);

	if (ft->fmtinfo.format == WIOS_PCM_FORMAT_S24) {
		if (err == nframes*n) {
			ft->success_byte_cnt += err > 0 ? (err*3)/4 : 0;
		} else {
			ft->success_byte_cnt += err > 0 ? (err*3)/4 : 0;
			ft->missing_byte_cnt += (nframes*n-err) > 0 ? (nframes*n-err*3)/4 : 0;
		}
	} else {
		if (err == nframes*n) {
			ft->success_byte_cnt += err > 0 ? err : 0;
		} else {
			ft->success_byte_cnt += err > 0 ? err : 0;
			ft->missing_byte_cnt += (nframes*n-err) > 0 ? (nframes*n-err) : 0;
		}
	}

	switch (ft->fmtinfo.format) {
		case WIOS_PCM_FORMAT_S16: {
			int16_t * buf1 = (int16_t *)buf->dv;
			for (i = 0; i < nframes*ft->fmtinfo.channels; i++) {
				buf->fl[i] = WIOS_CONVERT_16_TO_32BIT(buf1[i]);
				AD_CLIP_WARN(WIOS_16BIT_CLIP_COUNT(buf1[i], ft->clips));
				WIOS_RENEW_MIN_MAX_FRAME(buf1[i], ft->min_item_val, ft->max_item_val);
			}
			break;
		}
		case WIOS_PCM_FORMAT_S24: {
			int24_t * buf1 = (int24_t *)buf->dv;
			for (i = 0; i < nframes*ft->fmtinfo.channels; i++) {
				buf->fl[i] = WIOS_CONVERT_24_TO_32BIT(buf1[i]);
				AD_CLIP_WARN(WIOS_24BIT_CLIP_COUNT(buf1[i], ft->clips));
				WIOS_RENEW_MIN_MAX_FRAME(buf1[i], ft->min_item_val, ft->max_item_val);
			}
			break;
		}
		default: {
			break;
		}
	}

	return err/n;
}

char * rasp_buffer_malloc(wios_file_t * ft, wios_buffer_t * buf, int first_nframes)
{
	return wios_buffer_malloc(ft, buf, first_nframes);
}

/* カンマで区切られた数字文字列から配列に値を取得する */
static int get_numlist(int * arr, char * s, int channels)
{
	int ch = 0;
	const char c = ',';
	int len, i;
	char * p = s, * st = s;

	if (s == NULL)
		WIOS_ERROR_S("string of numbers is NULL");
	
	len = strlen(s);

	if (len > MAX_STRING_LENGTH)
		WIOS_ERROR_S("string of numbers is too long");

	/* arrに値を取得 */
	for (i = 0; i < channels; i++) {
		char buf[128];
		memset(buf, 0, sizeof(buf));
		p = strchr(p, c);
		if (p) {
			strncpy(buf, st, p-st);
			arr[ch++] = atoi(buf);
			st = ++p;
		} else {
			strncpy(buf, st, s+len-st);
			arr[ch++] = atoi(buf);
			break;
		}
	}

	return ch;

error:
	return WIOS_EOF;
}

/* DAゲインの値をチャンネルごとに設定 */
static int rasp_set_da_gain(wios_file_t * ft)
{
int da_gain[MAX_BOARDS];
	int ch, i;
	wios_fmtinfo_t * fmt = &ft->fmtinfo;

	if ((ch=get_numlist(da_gain, rasp_da_gain_str, fmt->channels)) == WIOS_EOF)
		goto error;
	
	if (ch != fmt->channels)
		WIOS_ERROR_S("the number of gain string's value is not equal to channels");
	
	/* ゲイン値のチェック */
	for (i = 0; i < ch; i++) {
		if (check_da_gain(da_gain[i]) == WIOS_EOF)
			WIOS_ERROR("DA gain value must be [-63, 0] (%d)", da_gain[i]);
	}

	/* ゲインを各チャンネルに設定 */
	for (i = 0; i < ch; i++) {
		/* WsSetDaGain関数の戻り値はリファレンスには正常時に0を返すとあるが、
		 * サンプルプログラムで確認したところ正常時に1を返している。 バグと思われる。*/
		// FIXME デバイスファームが上がったので要変更
		if (WsSetDaGain(ft->dp, i, da_gain[i]) != 0)
			WIOS_ERROR("cannot set DA gain %d to channel %d", da_gain[i], i+1);
		//pp(i, da_gain[i]);
	}

	/*for (i = 0; i < ch; i++) {
		int v;
		p(WsGetDaGain(ft->dp, i, &v));
		pp(i,v);
	}*/

	return WIOS_SUCCESS;

error:
		return WIOS_EOF;
}

/* ADゲインの値をチャンネルごとに設定 */
static int rasp_set_ad_gain(wios_file_t * ft)
{
	int ad_gain[MAX_BOARDS*CHANNELS_PER_BOARD];
	int ch, i;
	wios_fmtinfo_t * fmt = &ft->fmtinfo;

	if ((ch = get_numlist(ad_gain, rasp_ad_gain_str, fmt->channels)) == WIOS_EOF)
		goto error;
	
	if (ch != fmt->channels)
		WIOS_ERROR_S("the number of gain string's value is not equal to channels");
	
	/* ゲイン値のチェック */
	for (i = 0; i < ch; i++) {
		if (check_ad_gain(ad_gain[i]) == WIOS_EOF)
			WIOS_ERROR("AD gain value must be 0, [10, 65] (%d)", ad_gain[i]);
	}

	/* ゲインを各チャンネルに設定 */
	for (i = 0; i < ch; i++) {
		if (WsSetCaptureGain(ft->dp, i, ad_gain[i]) != 0)
			WIOS_ERROR("cannot set AD gain %d to channel %d", ad_gain[i], i+1);
	}

	return WIOS_SUCCESS;

error:
		return WIOS_EOF;
}

/* DAゲインの値を全て同じ値に設定 */
static int rasp_set_da_gain_all(wios_file_t * ft)
{
	int ch, i;
	wios_fmtinfo_t * fmt = &ft->fmtinfo;
	ch = fmt->channels;
	
	/* ゲイン値のチェック */
	if (check_da_gain(rasp_da_gain_val) == WIOS_EOF)
		WIOS_ERROR("DA gain value must be [%d, %d] (not %d)", 
				DA_GAIN_MIN, DA_GAIN_MAX, rasp_da_gain_val);

	/* ゲインを各チャンネルに設定 */
	for (i = 0; i < ch; i++) {
		if (WsSetDaGain(ft->dp, i, rasp_da_gain_val) != 0)
			WIOS_ERROR("cannot set DA gain %d to channel %d", rasp_da_gain_val, i+1);
	}

	return WIOS_SUCCESS;

error:
	return WIOS_EOF;
}

/* ADゲインの値を全て同じ値に設定 */
static int rasp_set_ad_gain_all(wios_file_t * ft)
{
	int ch, i;
	wios_fmtinfo_t * fmt = &ft->fmtinfo;
	ch = fmt->channels;
	
	/* ゲイン値のチェック */
	if (check_ad_gain(rasp_ad_gain_val) == WIOS_EOF)
		WIOS_ERROR("AD gain value must be [%d, %d] (not %d)", 
				AD_GAIN_MIN, AD_GAIN_MAX, rasp_ad_gain_val);

	/* ゲインを各チャンネルに設定 */
	for (i = 0; i < ch; i++) {
		if (WsSetCaptureGain(ft->dp, i, rasp_ad_gain_val) != 0)
			WIOS_ERROR("cannot set AD gain %d to channel %d", rasp_ad_gain_val, i+1);
	}

	return WIOS_SUCCESS;

error:
	return WIOS_EOF;
}

// TODO デバイス更新後要確認
int rasp_set_loopback_ch(wios_file_t * ft)
{
	int da_lb_ch[MAX_BOARDS*CHANNELS_PER_BOARD];
	int ad_lb_ch[MAX_BOARDS*CHANNELS_PER_BOARD];
	int da_channels, ad_channels, i;

	da_channels = get_numlist(da_lb_ch, rasp_da_lb_ch_str, MAX_BOARDS*CHANNELS_PER_BOARD);
	ad_channels = get_numlist(ad_lb_ch, rasp_ad_lb_ch_str, MAX_BOARDS*CHANNELS_PER_BOARD);

	if (da_channels == WIOS_EOF || ad_channels == WIOS_EOF)
		goto error;
	if (da_channels != ad_channels)
		WIOS_ERROR_S("the number of channel's in DA and AD list is not same");
	
	/* ループバックする各チャンネルを設定 */
	for (i = 0; i < da_channels; i++) {
		pp(da_lb_ch[i], ad_lb_ch[i]);
		if (WsSetDaLoopBackChannel(ft->dp, da_lb_ch[i], ad_lb_ch[i]) != 0)
			WIOS_ERROR("cannot set loopback channels DA:`%d' AD:`%d'", da_lb_ch[i], ad_lb_ch[i]);
	}

	/* DAの各チャンネルの出力モードの設定 */
	for (i = 0; i < da_channels; i++) {
		if (WsSetDaMode(ft->dp, da_lb_ch[i], DA_LOOP_BACK_MODE) != 0)
			WIOS_ERROR("cannot set ch:%d to DA loopback mode", da_lb_ch[i]);
	}
	
	return WIOS_SUCCESS;
	
error:
	return WIOS_EOF;
}
// TODO


/* GLOBAL FUNCTION */
/* -------------------------------------------------- */

wios_format_handler_t const * wios_set_rasp_format_handler(void)
{
	static wios_format_handler_t const handler = {
		device_init,
		open_read, open_write,
		playback, capture,
		finish
	};

	return &handler;
}
